package lpll;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Random;
import lpll.lplist;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.core.converters.ConverterUtils.DataSource;
import weka.filters.Filter;
import weka.filters.unsupervised.instance.Resample;

public class Cross_Validation {
	//the lp-structures
    static lplist lpList;
    static double []resultsCINdex;
    //monotony indicator
    static int [] monotony;


	public static void main(String[] args) throws Exception {
	
		 //load and prepare the data 
		 String location = "C:\\01_blood_transfusion"; //USER INPUT
		 // Read all the instances in the file (ARFF)
		 String file = "\\transfusion.arff"; //USER INPUT
		 DataSource source = new DataSource(location + file);
		 Instances orgData = source.getDataSet(); //assuming that there neither missing nor negative attribute values otherwise preprocessing steps are necessary
		 
		 //if desired consider only a sub-sample of the pairwise preferences for determining r* and g* if necessary (depending on available performance)
		 double proportion = 100;	//USER INPUT; e.g. for classification data with more than 500 instances 20 % should be sufficient to determine r* and g*
		 //delete duplicates
		 orgData = duplicates(orgData);
		 
		 //parameters w.r.t. LPLL
		 //monotony indicator to consider background knowlege if available
		 monotony = new int[orgData.numAttributes()]; 
		 for(int i = 0; i < orgData.numAttributes()-1; i++){
			 //set monotony indicator per default 0 -> no monotony for attribute i
			 monotony[i] = 0;
		 }
		 /*monotony[0] = 1; 	//USER INPUT; 1 indicating increasing and 2 decreasing order of attribute values
		 monotony[1] = 1;
		 monotony[2] = 1;
		 monotony[3] = 1;
		 monotony[4] = 1;
		 monotony[5] = 1;*/
		 
		 //parameters w.r.t. the cross validation and determination of hyper parameters
		 //set maximum levels grouping and binning parameter 
		 int maxGroup = 3;	// USER INPUT
		 int maxBins = 3; 	//USER INPUT; 3 corresponds to 4 bins etc.
		 //the number of folds
		 int folds = 10;
		 //the number of repetitions 
		 int numberRep = 10;
		 //the seed for randomizing the data
		 int seed = 23;
		 
		 //randomize data
		 Random rand = new Random(seed);
		 orgData.randomize(rand);
		 
		 //call nested cross validation
		 nestedVali(proportion, orgData, location, folds, numberRep, maxGroup, maxBins);
	    
	}

	
	
	/**
     * Nested cross-validation to 1) determine hyper parameters 2) validate with the parameters determined
     * @param proportion The proportion to be considered to shrink the dataset if necessary
     * @param orgData The data set
     * @param path The path for saving the results
     * @param dataset Additional information for the path
     * @param folds The amount of folders
     * @param repeat The number of repetitions
     * @param maxGroup The maximal grouping size for CLeRa
     * @param maxBins The maximum number of bins for discretization of attribute values
     */
    public static void nestedVali(double proportion, Instances orgData, String dataset, int folds, int repeat, int maxGroup, int maxBins) throws Exception {
   	 	int seed = 23;
   	 	//create seeded number generator
        Random rand = new Random(seed);
       
        //the results
        double [][] results = new double[folds][];
        double [][] repResults = new double [repeat][];
        double repParamR[] = new double [repeat];
        double repParamG[] = new double [repeat];
        double paramR[] = new double [folds];
        double paramG[] = new double [folds];

        //nested cross validation for determination of hyperparameters
        for(int rep = 0; rep < repeat; rep++){
        	for (int runs = 0; runs < folds; runs++){
        		
        		//randomize the data for every loop 
        		seed = seed *(runs+1);
        		rand = new Random(seed);
        		orgData.randomize(rand);
        		//i. Split the sample S into n chunks
        		Instances train = orgData.trainCV(folds, runs);
        		//data for outer cross-validation
        		Instances test = orgData.testCV(folds, runs);

        		//initiate parameters
                int bestbin = 1;
                int bestgroup = 1;
                double bestResultBin = 0;
                double avlpListCorrectnessTest = 0;
                double avlpListCompletnessTest = 0;
                int errorsLPList[];
                
                //ii. For all possible parameter settings which have to be tried out, invoke the learner, and evaluate the resulting hypothesis on the remaining chunk of Si which has not been used for training. Thus, determine the parameter setting which leads to the least error on the hold-out chunk.
       			for(int i = 1; i <= maxBins; i++){
   		    		int bins = i;
      		    	for(int group = 1; group <= maxGroup; group++){
      		    		double criterium = 0;
      		    		for(int k = 0; k < folds; k++){
      		    			rand = new Random(seed);
      		    			orgData.randomize(rand);
      		    			//Split Si into n chunks and repeat n times 
      		    			Instances trainInner = train.trainCV(folds, k);
      		    			//data for outer cross-vali
      		    			Instances testInner = train.testCV(folds, k);
      		    			//create set of pairwise preferences
      		    			Instances trainPW = transformData(trainInner);
      		    			trainPW = resample(proportion, trainPW);
      		    			Instances testPW = transformData(testInner);
      		    				
      		    			//determine the lp list
      		    			lplist liste = new lplist(trainPW, group, bins, monotony);
      		    				
      		    			//calculate corrrectness and completeness for LP-List (based on test data)
      		    			errorsLPList = liste.computeErrorsExt(testPW);
      		    			double lpListCorrectnessTest = (testPW.numInstances()-errorsLPList[1]-2*errorsLPList[0]);
      		    			lpListCorrectnessTest = lpListCorrectnessTest / (testPW.numInstances()-errorsLPList[1]);
      		    			double lpListCompletnessTest = (testPW.numInstances()-errorsLPList[1]);
      		    			lpListCompletnessTest = lpListCompletnessTest/testPW.numInstances();
      		    			avlpListCorrectnessTest = avlpListCorrectnessTest + lpListCorrectnessTest;
      		    			avlpListCompletnessTest = avlpListCompletnessTest + lpListCompletnessTest;
      		    		}
      		    		//calculate average values
      		    		avlpListCorrectnessTest = avlpListCorrectnessTest/folds;
      		    		avlpListCompletnessTest = avlpListCompletnessTest/folds;
      		    		criterium = avlpListCorrectnessTest * avlpListCompletnessTest;
      		    		if(criterium > bestResultBin){	
                    		bestResultBin = criterium;
                    		bestbin = i;
                    		bestgroup = group;
                    	}
      		    		avlpListCorrectnessTest = 0;
      		    		avlpListCompletnessTest = 0; 
      		    	}
       			}
       			    
                //iii. Use the parameter setting determined by the parameter optimizer and invoke the learner, providing it with Si . 
       			double tmpresults[] =  validateLpList(train, test, bestgroup, bestbin);
       			results[runs] = new double[5];
       			results[runs][0] = tmpresults[0];
       			results[runs][1] = tmpresults[1];
       			results[runs][2] = tmpresults[2];
       			results[runs][3] = tmpresults[3];
       			results[runs][4] = tmpresults[4];
       			paramR[runs] =bestbin;
       			paramG[runs] = bestgroup;
        	}
        
        	double avLpListCInd = 0; 
	        double stdLpListCInd = 0;
	        double avR = 0;
	        double avG = 0;
	               
	        //compute average indices
	        for(int k = 0; k < folds; k++){
	        	avLpListCInd = avLpListCInd + results[k][0];
	            avR = avR+ paramR[k];
	            avG = avG+ paramG[k];
	        }
	        avLpListCInd = avLpListCInd/folds;
	        avR = avR/folds;
	        avG = avG/folds;
               
	        //compute standard deviation
	        for(int k = 0; k < folds; k++){
	        	stdLpListCInd = stdLpListCInd + (results[k][0] - avLpListCInd)*(results[k][0] - avLpListCInd);
	        }
                    	 
	        stdLpListCInd = Math.sqrt(stdLpListCInd / folds);
	             	  
	        //return average cIndices and standard deviation :
	        System.out.println("Results for "+folds +"-fold cross validation: ");
	        System.out.println("Average cIndex for LP-List: " + avLpListCInd);
	        System.out.println("Standard deviation: " + stdLpListCInd);
	        System.out.println("Average best binning r*: " + (avR+1));
	        System.out.println("Average best grouping g*: " + avG);
	       	
	        if(repeat > 1){
	   			repResults[rep] = new double[6];
	   			repResults[rep][2] = avLpListCInd;
	   			repResults[rep][3] = stdLpListCInd;
	   			repParamR[rep] = avR;
	   			repParamG[rep] = avG;
	   		}
        }
       
        //return repeated cross validation results
        if(repeat > 1){
       		double sumCind = 0;
       		double sumCindSdt = 0;
       		double sumAttr = 0;
       		double sumAttrSdt = 0;
       		double sumR= 0;
       		double sumG = 0;
       		for(int i = 0; i < repResults.length; i++){
       			//plausibility check //added 06072016
       			if(repResults[i][2] > 0 && repParamR[i] > 0 && repParamG[i] >0 ){
       				sumCind = sumCind +repResults[i][2];
           			sumCindSdt = sumCindSdt +repResults[i][3];
           			sumAttr = sumAttr +repResults[i][4];
           			sumAttrSdt = sumAttrSdt +repResults[i][5];
           			sumR = sumR + repParamR[i];
           			sumG = sumG + repParamG[i];	
       			}
       			else{
       				repeat = repeat -1;
       			}
       			
       		}
       		sumCind = sumCind / repeat;
       		sumCindSdt = sumCindSdt / repeat;
       		sumAttr = sumAttr / repeat;
       		sumAttrSdt = sumAttrSdt / repeat;
       		sumR = sumR / repeat;
       		sumG = sumG / repeat;
       		
       		System.out.println("Results for " +repeat +" - " +folds +" cross-validation: ");
       		System.out.println("Average cIndex for LP-List: " + sumCind);
          	System.out.println("Standard deviation: " + sumCindSdt);
          	//in case that there are nominal attributes report to user
          	for(int k = 0; k < orgData.numAttributes(); k++){
          		if(orgData.attribute(k).isNominal()){
          			System.out.println("Nominal attribute values: Please consider that the maximum number of bins might be higher than r* as nominal attributes do not require discretization");
          			break;
          		}
          	}
          	System.out.println("Average best binning r*: " + (sumR+1));
          	System.out.println("Average best grouping g*: " + sumG);
        }
       	
    }

    
    
    /**
     * Learn and evaluate the lp-list given a parametrization w.r.t. g and r
     * @param 	train The training data set
     * @param 	test The training data set
     * @param 	group The grouping size for learning the lp-list
     * @param 	bins The number of bins for discretization of attribute values
     * @return 	Performance measures for the resulting lp-list given g and r
     */
    public static double[] validateLpList(Instances train, Instances test, int group, int bins) throws Exception{
   	 	//Performance measures
        double lpListCIndex;
        double lpListCIndexTrain;
        //vector storing the results: [0] = cIndex(test), [1]  = cIndex(train);
        double result[] = new double[5];

        //create set of pairwise preferences
        Instances trainPW = transformData(train);
               	 
        //learn the LP-List
        lpList = new lplist(trainPW, group, bins, monotony);
        //compute corrrectness and completeness for LP-List (based on training data)
        lpList.computeErrorsExt(trainPW);
        //Rank the test data set by using the LP-List
        rankingLPList(test);
        //Rank the training data using the LP-List
        rankingLPList(train);
                     	 
        //Evaluate the ranked data (in terms of C-index)
        lpListCIndex = cIndexWithAbstention(test);
        lpListCIndexTrain = cIndexWithAbstention(train);
           
        result[0] = lpListCIndex;
        result[1] = lpListCIndexTrain;
        return result;
          	 
    }
        
    
    
    /**
     * Ranks the objects based on the LP-List using insertion sort 
     * @param data The data to rank
     * @return Ranked data according to the LP list
     */
    public static Instances rankingLPList(Instances data) throws Exception{
        
   	 	//Instances to be compared during insertion sort
        Instance one = new DenseInstance(data.instance(0).numAttributes()-1);
        Instance two = new DenseInstance(data.instance(0).numAttributes()-1);

        int j = 0;
        for(int i = 1; i < data.numInstances(); i++){
            one = copyInst(data.instance(i-1));
            two = copyInst(data.instance(i));
            j = i;
            while(j > 0 && lpList.decideExt(one, two, data) != 1){
                data.swap(j-1, j);
                j = j-1;
                if(j>0){
                    one = copyInst(data.instance(j-1));
                    two = copyInst(data.instance(j));
                }
            }
        }
               
        return data;
    }
    
    
    
    
    /**
     * Computes the C-Index for the given, ranked data-set (considering abstention/ties)
     * @param data The data set for evaluation
     * @return The C-Index
     */
    public static double cIndexWithAbstention(Instances data) throws Exception{
        int classVal = 0;
        double correct = 0;
        double abstention = 0;
        double result;

        Instance one = new DenseInstance(data.instance(0).numAttributes());
        Instance two = new DenseInstance(data.instance(0).numAttributes());
        data.setClassIndex(data.numAttributes()-1);
        
        //Compute the number of correct by identifying pairs of objects with different class values
        for(int i = 0; i < data.numInstances(); i++){
        	
        	classVal = (int) data.instance(i).value(data.numAttributes()-1);
	        one = copyInst(data.instance(i));
	        for(int j = 0; j < data.numInstances();j++){
	        	if(j != i){
	           		if(data.instance(j).value(data.numAttributes()-1) < classVal){
	           			two = copyInst(data.instance(j));
	           			int erg = lpList.decideExt(one, two, data);
	           			//abstention
	           			if( erg == 0){
	           				abstention++;
	           			}
	           			//objects were correctly ordered
	           			else if (erg == 1){
	           				correct++;
	           			}	
	           		 }
	           		one = copyInst(data.instance(i));
	           	}
	        }
        }
     	  
        //compute the sum of the product of different classes
        double sum = 0;
        int[][] classValues = classValsDist(data);
        for(int i = 0; i < classValues.length; i++){
            for(int j = i+1 ; j < classValues.length; j++){
                sum = sum + classValues[i][1]*classValues[j][1];
            }
        }
        result = (correct+ 0.5 * abstention)/sum;
        return result;
    }
    
    
    
	/**
	 * Transform a classification data set into a set of pairwise preferences
	 * @param input The data to be transformed
	 * @return The respective set of pairwise preferences
	 */
	public static Instances transformData(Instances input) throws Exception{
		 //Creating the attributes for the output data
		 ArrayList<Attribute> attInfo = new ArrayList<Attribute>((input.numAttributes()-1)*2);
		 int i = 0;
		 for (int j = 0; j < (input.numAttributes()-1)*2; j++){
			 if(i == (input.numAttributes()-1)){
				 i = 0;
			 }
			 if(i < input.numAttributes() && input.attribute(i).isNominal()){
				 ArrayList<String> nominal_values = new ArrayList<String>(input.attribute(i).numValues()); 
				 Enumeration enu= input.attribute(i).enumerateValues();
				 while(enu.hasMoreElements()) {
					 nominal_values.add((String) enu.nextElement());
				 }
				 Attribute new_nominal = new Attribute("A"+(j+1), nominal_values);
				 attInfo.add(new_nominal );	
			 }
			 else{
				 attInfo.add(new Attribute("A"+(j+1)));	 
			 }
			 i++;
		 }
		 DenseInstance pair = new DenseInstance((input.numAttributes()-1)*2);
	    
		 //generate the corresponding set of pairwise preferences w.r.t. the class values
	     Instances output = new Instances("Pairwise preferences",attInfo, input.numInstances()*2);
	     int classValue;
	     for(i = 0; i < input.numInstances(); i++){
	   	  	classValue = (int) input.instance(i).value(input.numAttributes()-1);
	        //compare the instance's class value to the other instances' class values
	        for(int j = 0; j < input.numInstances();j++){
	        	if(j != i && (classValue > input.instance(j).value(input.numAttributes()-1))){
		       		//outcome one
		            for(int k = 0; k < input.numAttributes()-1; k++){
		            	pair.setValue(k, input.instance(i).value(k));
		            }
		            //outcome two
		            for(int k = 0; k < input.numAttributes()-1; k++){
		            	pair.setValue(k+(input.numAttributes()-1),input.instance(j).value(k));
		            }
		            output.add(pair);
	            }
	        }
	
	     }
	     return output;
	}
	
	
	
	/**
     * Resample and resize the given data set
     * @param proportion The proportion of the orginial data-set
     * @param dataset The dataset to be resampled
     * @return The respective data set
     */
    public static Instances resample(double proportion, Instances original_train) throws Exception {
   	 	//original_train.setClassIndex(original_train.numAttributes()-1);
        if(proportion != 100.0){
        	Resample resample = new Resample();
        	resample.setRandomSeed(23);
        	resample.setInputFormat(original_train);
        	resample.setSampleSizePercent(proportion);
        	return Filter.useFilter(original_train, resample);}
        else{
       	 	return original_train;
        }
    }
	
    
    
    /**
     * Returns a copy of the given instance
     * @param inst The instance to be copied
     * @return A copy of the instance
     */
    public static Instance copyInst(Instance inst){
        Instance copy = new DenseInstance(inst.numAttributes());
        for(int i = 0; i < copy.numAttributes();i++){
            copy.setValue(i, inst.value(i));
        }
        return copy;
    }
    
    
    
    /**
     * Returns the distinct values of the class attribute in the set and their respective occurrence
     *@param data The data for exploration
     *@return Array filled with the labels for given attribute and their occurrence
     */
     public static int[][] classValsDist(Instances data){
    	 int result[][];
    	 int [] distVals = new int[data.numDistinctValues(data.numAttributes()-1)];
    	 double [] attVals = data.attributeToDoubleArray(data.numAttributes()-1);

         //Position of the ascending ordered instances with respect to their attribute value
         int [] sorted = Utils.sort(attVals);
         double prev = 0;
         int counter = 0;

         //Identify distinct attribute values
         for (int i = 0; i < sorted.length; i++) {
        	 Instance current = data.instance(sorted[i]);
             if (current.isMissing(data.numAttributes()-1)) {
            	 break;
             }
             if ((i == 0) || (current.value(data.numAttributes()-1) > prev)) {
            	 prev = current.value(data.numAttributes()-1);
                 distVals[counter]=(int)current.value(data.numAttributes()-1);
                 counter++;
             }
         }

         //determine the number of occurrence
         result = new int[distVals.length][2];
         for(int i = 0; i < distVals.length; i++){
        	 result[i][0] = distVals[i];
             result[i][1] = 0;
             for(int j = 0; j < data.numInstances(); j++){
            	 if((int) data.instance(j).value(data.numAttributes()-1) == result[i][0]){
            		 result[i][1]++;
            	 }
             }

         }
         return result;
     }
       
     
     
     
     /**
      * Identifies duplicates within the data and returns a revised dataset
      *@param data The data for exploration
      *@return The revised dataset without duplicates
      */
      public static Instances duplicates(Instances data){
    	 Instance one = new DenseInstance(data.numAttributes());
   	   	 Instance two = new DenseInstance(data.numAttributes());
   	   	 for(int i = 0; i < data.numInstances();i++){
   	   		 one = copyInst(data.instance(i));
   	   		 for(int j = 0; j < data.numInstances(); j++){
   	   			 if(j != i){
   	   				 two = copyInst(data.instance(j));
   	   				 boolean same = true;
   	   				 //exclude the class attribute within the comparison -> -1
   	   				 for(int k =0; k < one.numAttributes() -1; k++){
   	   					 if(one.value(k) != two.value(k)){
   	   						 same = false;
   	   					 }
   	   				 }
   	   				 //same instances --> delete instance two from the data set!
   	   				 if(same == true){
   	   					 data.delete(j);
   	   				 }
   	   			 }
   	   		 }
   	   	 }
   	   	 return data;
      }
      
      
      
    
}